@@ -35,6 +35,19 @@
 
 #include "openconnect.h"
 
+static unsigned char nybble(unsigned char n)
+{
+	if      (n >= '0' && n <= '9') return n - '0';
+	else if (n >= 'A' && n <= 'F') return n - ('A' - 10);
+	else if (n >= 'a' && n <= 'f') return n - ('a' - 10);
+	return 0;
+}
+
+unsigned char unhex(const char *data)
+{
+	return (nybble(data[0]) << 4) | nybble(data[1]);
+}
+
 #ifdef SSL_F_DTLS1_CONNECT
 #if 0
 /*
@@ -90,19 +103,6 @@ int RAND_bytes(char *buf, int len)
  * their clients use anyway.
  */
 
-static unsigned char nybble(unsigned char n)
-{
-	if      (n >= '0' && n <= '9') return n - '0';
-	else if (n >= 'A' && n <= 'F') return n - ('A' - 10);
-	else if (n >= 'a' && n <= 'f') return n - ('a' - 10);
-	return 0;
-}
-
-static unsigned char hex(const char *data)
-{
-	return (nybble(data[0]) << 4) | nybble(data[1]);
-}
-
 int connect_dtls_socket(struct openconnect_info *vpninfo)
 {
 	STACK_OF(SSL_CIPHER) *ciphers;
@@ -211,7 +211,7 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 	if (!SSL_set_session(dtls_ssl, vpninfo->dtls_session)) {
 		vpninfo->progress(vpninfo, PRG_ERR,
 				  "SSL_set_session() failed with old protocol version 0x%x\n"
-				  "Your OpenSSL may lack Cisco compatibility support\n"
+				  "Are you using a version of OpenSSL older than 0.9.8m?\n"
 				  "See http://rt.openssl.org/Ticket/Display.html?id=1751\n"
 				  "Use the --no-dtls command line option to avoid this message\n",
 				  vpninfo->dtls_session->ssl_version);
@@ -346,7 +346,7 @@ int setup_dtls(struct openconnect_info *vpninfo)
 				return -EINVAL;
 			}
 			for (i = 0; i < 64; i += 2)
-				vpninfo->dtls_session_id[i/2] = hex(dtls_opt->value + i);
+				vpninfo->dtls_session_id[i/2] = unhex(dtls_opt->value + i);
 			sessid_found = 1;
 		} else if (!strcmp(dtls_opt->option + 7, "Port")) {
 			dtls_port = atol(dtls_opt->value);
