@@ -51,7 +51,8 @@ static int proxy_write(int fd, unsigned char *buf, size_t len);
  * provided by their caller.
  */
 
-int http_add_cookie(struct openconnect_info *vpninfo, const char *option, const char *value)
+static int http_add_cookie(struct openconnect_info *vpninfo,
+			   const char *option, const char *value)
 {
 	struct vpn_option *new, **this;
 
@@ -124,7 +125,8 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		return -EINVAL;
 	}
 
-	vpninfo->progress(vpninfo, PRG_TRACE, "Got HTTP response: %s\n", buf);
+	vpninfo->progress(vpninfo, (*result==200)?PRG_TRACE:PRG_INFO,
+			  "Got HTTP response: %s\n", buf);
 
 	/* Eat headers... */
 	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
@@ -289,7 +291,7 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		}
 	}
 
-	if (closeconn) {
+	if (closeconn || vpninfo->no_http_keepalive) {
 		SSL_free(vpninfo->https_ssl);
 		vpninfo->https_ssl = NULL;
 		close(vpninfo->ssl_fd);
@@ -365,10 +367,10 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 	int fd, ret;
 
 	if (!vpninfo->uid_csd_given) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Error: You are trying to "
-				  "run insecure CSD code without specifying the CSD user.\n"
-				  "       Use command line option \"--csd-user\"\n");
-		exit(1);
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Error: Server asked us to download and run a 'Cisco Secure Desktop' trojan.\n"
+				  "This facility is disabled by default for security reasons, so you may wish to enable it.");
+		return -EPERM;
 	}
 
 #ifndef __linux__
@@ -427,7 +429,11 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 				"CSD code with root privileges\n"
 				"\t Use command line option \"--csd-user\"\n");
 		}
-
+		if (vpninfo->uid_csd_given == 2) {	       
+			/* The NM tool really needs not to get spurious output
+			   on stdout, which the CSD trojan spews. */
+			dup2(2, 1);
+		}
 		csd_argv[i++] = fname;
 		csd_argv[i++] = "-ticket";
 		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket) == -1)
@@ -532,11 +538,8 @@ int parse_url(char *url, char **res_proto, char **res_host, int *res_port,
 	}
 
 	path = strchr(host, '/');
-	if (path) {
+	if (path)
 		*(path++) = 0;
-		if (!*path)
-			path = NULL;
-	}
 
 	port_str = strrchr(host, ':');
 	if (port_str) {
@@ -556,7 +559,13 @@ int parse_url(char *url, char **res_proto, char **res_host, int *res_port,
 	if (res_port)
 		*res_port = port;
 	if (res_path)
-		*res_path = path ? strdup(path) : NULL;
+		*res_path = (path && *path) ? strdup(path) : NULL;
+
+	/* Undo the damage we did to the original string */
+	if (path)
+		*(path - 1) = '/';
+	if (proto)
+		*(host - 3) = ':';
 	return 0;
 }
 
@@ -712,7 +721,13 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 			goto retry;
 		}
 	}
-
+	if (!form_buf || result != 200) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Unexpected %d result from server\n",
+				  result);
+		free(form_buf);
+		return -EINVAL;
+	}
 	if (vpninfo->csd_stuburl) {
 		/* This is the CSD stub script, which we now need to run */
 		result = run_csd_script(vpninfo, form_buf, buflen);
