@@ -37,8 +37,9 @@
 #include <sys/utsname.h>
 #include <sys/types.h>
 #include <openssl/rand.h>
+#include <openssl/ui.h>
 #ifdef OPENCONNECT_LIBPROXY
-#include <libproxy/proxy.h>
+#include LIBPROXY_HDR
 #endif
 
 #define _GNU_SOURCE
@@ -49,10 +50,12 @@
 static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buflen);
 static void write_progress(struct openconnect_info *info, int level, const char *fmt, ...);
 static void syslog_progress(struct openconnect_info *info, int level, const char *fmt, ...);
+static int validate_peer_cert(struct openconnect_info *info, X509 *peer_cert, const char *reason);
 
 int verbose = PRG_INFO;
 int background;
 int do_passphrase_from_fsid;
+int nocertcheck;
 
 static struct option long_options[] = {
 	{"background", 0, 0, 'b'},
@@ -95,6 +98,8 @@ static struct option long_options[] = {
 	{"disable-ipv6", 0, 0, 0x05},
 	{"no-proxy", 0, 0, 0x06},
 	{"libproxy", 0, 0, 0x07},
+	{"no-http-keepalive", 0, 0, 0x08},
+	{"no-cert-check", 0, 0, 0x09},
 	{NULL, 0, 0, 0},
 };
 
@@ -140,7 +145,9 @@ void usage(void)
 	printf("      --disable-ipv6              Do not ask for IPv6 connectivity\n");
 	printf("      --dtls-ciphers=LIST         OpenSSL ciphers to support for DTLS\n");
 	printf("      --no-dtls                   Disable DTLS\n");
+	printf("      --no-http-keepalive         Disable HTTP connection re-use\n");
 	printf("      --no-passwd                 Disable password/SecurID authentication\n");
+	printf("      --no-cert-check             Do not require server SSL cert to be valid\n");
 	printf("      --passwd-on-stdin           Read password from standard input\n");
 	printf("      --reconnect-timeout         Connection retry timeout in seconds\n");
 	printf("      --servercert                Server's certificate SHA1 fingerprint\n");
@@ -206,6 +213,7 @@ int main(int argc, char **argv)
 	vpninfo->reconnect_timeout = 300;
 	vpninfo->uid_csd = 0;
 	vpninfo->uid_csd_given = 0;
+	vpninfo->validate_peer_cert = validate_peer_cert;
 
 	if (RAND_bytes(vpninfo->dtls_secret, sizeof(vpninfo->dtls_secret)) != 1) {
 		fprintf(stderr, "Failed to initialise DTLS secret\n");
@@ -327,6 +335,14 @@ int main(int argc, char **argv)
 			autoproxy = 1;
 			proxy = NULL;
 			break;
+		case 0x08:
+			fprintf(stderr, "Disabling all HTTP connection re-use due to --no-http-keepalive option.\n"
+				"If this helps, please report to <openconnect-devel@lists.infradead.org>.\n");
+			vpninfo->no_http_keepalive = 1;
+			break;
+		case 0x09:
+			nocertcheck = 1;
+			break;
 		case 's':
 			vpninfo->vpnc_script = optarg;
 			break;
@@ -501,6 +517,10 @@ int main(int argc, char **argv)
 			      (vpninfo->deflate ? "SSL + deflate" : "SSL")
 			      : "DTLS");
 
+	if (!vpninfo->vpnc_script)
+		vpninfo->progress(vpninfo, PRG_INFO,
+				  "No --script argument provided; DNS and routing are not configured\n");
+
 	if (background) {
 		int pid;
 		if ((pid = fork())) {
@@ -563,3 +583,68 @@ void syslog_progress(struct openconnect_info *info, int level,
 		va_end(args);
 	}
 }
+
+
+struct accepted_cert {
+	struct accepted_cert *next;
+	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
+	char host[0];
+} *accepted_certs;
+
+static int validate_peer_cert(struct openconnect_info *vpninfo, X509 *peer_cert,
+			      const char *reason)
+{
+	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
+	struct accepted_cert *this;
+	int ret;
+
+	if (nocertcheck)
+		return 0;
+
+	ret = get_cert_sha1_fingerprint(vpninfo, peer_cert, fingerprint);
+	if (ret)
+		return ret;
+
+	for (this = accepted_certs; this; this = this->next) {
+		if (!strcasecmp(this->host, vpninfo->hostname) &&
+		    !strcasecmp(this->fingerprint, fingerprint))
+			return 0;
+	}
+	
+	while (1) {
+		UI *ui;
+		char buf[6];
+		int ret;
+
+		fprintf(stderr, "\nCertificate from VPN server \"%s\" failed verification.\n"
+			"Reason: %s\n",	vpninfo->hostname, reason);
+		fflush(stderr);
+
+		ui = UI_new();
+		UI_add_input_string(ui, "Enter 'yes' to accept, 'no' to abort; anything else to view: ",
+				    UI_INPUT_FLAG_ECHO, buf, 2, 5);
+		ret = UI_process(ui);
+		UI_free(ui);
+		if (ret == -2)
+			return -EINVAL;
+		if (ret == -1)
+			buf[0] = 0;
+
+		if (!strcasecmp(buf, "yes")) {
+			struct accepted_cert *newcert = malloc(sizeof(*newcert) +
+							       strlen(vpninfo->hostname) + 1);
+			if (newcert) {
+				newcert->next = accepted_certs;
+				accepted_certs = newcert;
+				strcpy(newcert->fingerprint, fingerprint);
+				strcpy(newcert->host, vpninfo->hostname);
+			}
+			return 0;
+		}
+		if (!strcasecmp(buf, "no"))
+			return -EINVAL;
+
+		X509_print_fp(stderr, peer_cert);
+	}
+				
+}
