@@ -119,7 +119,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 			continue;
 
 		vpninfo->progress(vpninfo, PRG_TRACE,
-				  "Did no work; sleeping for %d ms...\n", timeout);
+				  "No work to do; sleeping for %d ms...\n", timeout);
 		memcpy(&rfds, &vpninfo->select_rfds, sizeof(rfds));
 		memcpy(&wfds, &vpninfo->select_wfds, sizeof(wfds));
 		memcpy(&efds, &vpninfo->select_efds, sizeof(efds));
