@@ -616,6 +616,7 @@ static char* get_title(const char *vpn_name)
 typedef struct cert_data {
 	auth_ui_data *ui_data;
 	X509 *peer_cert;
+	const char *reason;
 } cert_data;
 
 
@@ -637,8 +638,9 @@ static gboolean user_validate_cert(cert_data *data)
 	BIO_get_mem_ptr(bp, &certinfo);
 
 	title = get_title(data->ui_data->vpninfo->vpn_name);
-	msg = g_strdup_printf("Unknown certificate from VPN server \"%s\".\n"
-			      "Do you want to accept it?", data->ui_data->vpninfo->hostname);
+	msg = g_strdup_printf("Certificate from VPN server \"%s\" failed verification.\n"
+			      "Reason: %s\nDo you want to accept it?",
+			      data->ui_data->vpninfo->hostname, data->reason);
 
 	dlg = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION,
 				     GTK_BUTTONS_OK_CANCEL,
@@ -683,7 +685,7 @@ static gboolean user_validate_cert(cert_data *data)
 
 /* runs in worker thread */
 static int validate_peer_cert(struct openconnect_info *vpninfo,
-			      X509 *peer_cert)
+			      X509 *peer_cert, const char *reason)
 {
 	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
 	char *certs_data;
@@ -714,6 +716,7 @@ static int validate_peer_cert(struct openconnect_info *vpninfo,
 	data = g_slice_new(cert_data);
 	data->ui_data = ui_data; /* FIXME uses global */
 	data->peer_cert = peer_cert;
+	data->reason = reason;
 
 	g_mutex_lock(ui_data->form_mutex);
 
@@ -884,10 +887,11 @@ static int parse_xmlconfig(char *xmlconfig)
 
 static int get_config(char *vpn_uuid, struct openconnect_info *vpninfo)
 {
-	char *authtype;
+	char *proxy;
 	char *xmlconfig;
 	char *hostname;
 	char *group;
+	char *csd;
 	char *pem_passphrase_fsid;
 
 	gcl = gconf_client_get_default();
@@ -950,37 +954,25 @@ if (0) {
 
 	vpninfo->cafile = get_gconf_setting(gcl, config_path, NM_OPENCONNECT_KEY_CACERT);
 
-	authtype = get_gconf_setting(gcl, config_path, NM_OPENCONNECT_KEY_AUTHTYPE);
-	if (!authtype) {
-		fprintf(stderr, "No authentication type configured\n");
-		return -EINVAL;
+	csd = get_gconf_setting(gcl, config_path, "enable_csd_trojan");
+	if (csd && !strcmp(csd, "yes")) {
+		/* We're not running as root; we can't setuid(). */
+		vpninfo->uid_csd = getuid();
+		vpninfo->uid_csd_given = 2;
 	}
+	g_free(csd);
 
-	if (!strcmp(authtype, NM_OPENCONNECT_AUTHTYPE_PASSWORD)) {
-		vpninfo->username = get_gconf_setting(gcl, config_path,
-						      NM_OPENCONNECT_KEY_USERNAME);
-		return 0;
-	}
-	if (!strcmp(authtype, NM_OPENCONNECT_AUTHTYPE_CERT_TPM))
-		vpninfo->cert_type = CERT_TYPE_TPM;
-	else if (strcmp(authtype, NM_OPENCONNECT_AUTHTYPE_CERT)) {
-		fprintf(stderr, "Unknown authentication type '%s'\n", authtype);
+	proxy = get_gconf_setting(gcl, config_path, "proxy");
+	if (proxy && proxy[0] && set_http_proxy(vpninfo, proxy))
 		return -EINVAL;
-	}
 
-	/* It's a certificate */
 	vpninfo->cert = get_gconf_setting(gcl, config_path, NM_OPENCONNECT_KEY_USERCERT);
-	if (!vpninfo->cert) {
-		fprintf(stderr, "No user certificate configured\n");
-		return -EINVAL;
-	}
-
 	vpninfo->sslkey = get_gconf_setting(gcl, config_path, NM_OPENCONNECT_KEY_PRIVKEY);
 	if (!vpninfo->sslkey)
 		vpninfo->sslkey = vpninfo->cert;
 
 	pem_passphrase_fsid = get_gconf_setting(gcl, config_path, "pem_passphrase_fsid");
-	if (pem_passphrase_fsid && !strcmp(pem_passphrase_fsid, "yes"))
+	if (pem_passphrase_fsid && vpninfo->sslkey && !strcmp(pem_passphrase_fsid, "yes"))
 		passphrase_from_fsid(vpninfo);
 	g_free(pem_passphrase_fsid);
 				    
