@@ -35,7 +35,7 @@
 #include <sys/types.h>
 #include <unistd.h>
 #ifdef OPENCONNECT_LIBPROXY
-#include <libproxy/proxy.h>
+#include LIBPROXY_HDR
 #endif
 
 
@@ -173,6 +173,7 @@ struct openconnect_info {
 	char *dtls_ciphers;
 	uid_t uid_csd;
 	int uid_csd_given;
+	int no_http_keepalive;
 
 	char *cookie;
 	struct vpn_option *cookies;
@@ -210,6 +211,7 @@ struct openconnect_info {
 	char *ifname;
 
 	int mtu;
+	const char *banner;
 	const char *vpn_addr;
 	const char *vpn_netmask;
 	const char *vpn_addr6;
@@ -249,7 +251,7 @@ struct openconnect_info {
 
 	char *quit_reason;
 
-	int (*validate_peer_cert) (struct openconnect_info *vpninfo, X509 *cert);
+	int (*validate_peer_cert) (struct openconnect_info *vpninfo, X509 *cert, const char *reason);
 	int (*write_new_config) (struct openconnect_info *vpninfo, char *buf, int buflen);
 	int (*process_auth_form) (struct openconnect_info *vpninfo, struct oc_auth_form *form);
 
@@ -273,7 +275,7 @@ struct openconnect_info {
 #define AC_PKT_TERM_SERVER	9	/* Server kick */
 
 /* Ick */
-#if OPENSSL_VERSION_NUMBER >= 0x10000000L
+#if OPENSSL_VERSION_NUMBER >= 0x00909000L
 #define method_const const
 #else
 #define method_const
@@ -287,6 +289,7 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
 void shutdown_tun(struct openconnect_info *vpninfo);
 
 /* dtls.c */
+unsigned char unhex(const char *data);
 int setup_dtls(struct openconnect_info *vpninfo);
 int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout);
 int dtls_try_handshake(struct openconnect_info *vpninfo);
