@@ -31,12 +31,14 @@
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
 #if defined(__linux__)
 #include <sys/vfs.h>
-#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
+#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__OpenBSD__) || defined(__APPLE__)
 #include <sys/param.h>
 #include <sys/mount.h>
-#elif defined (__sun__)
+#elif defined (__sun__) || defined(__NetBSD__) || defined(__DragonFly__)
 #include <sys/statvfs.h>
 #endif
 
@@ -138,18 +140,35 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 {
 	EVP_PKEY *pkey = NULL;
 	X509 *cert = NULL;
-	STACK_OF(X509) *ca = sk_X509_new_null();
+	STACK_OF(X509) *ca;
 	int ret = 0;
 	char pass[PEM_BUFSIZE];
 
+ retrypass:
+	/* We do this every time round the loop, to work around a bug in
+	   OpenSSL < 1.0.0-beta2 -- where the stack at *ca will be freed
+	   when PKCS12_parse() returns an error, but *ca is left pointing
+	   to the freed memory. */
+	ca = NULL;
 	if (!vpninfo->cert_password) {
 		if (EVP_read_pw_string(pass, PEM_BUFSIZE,
 				       "Enter PKCS#12 pass phrase:", 0))
 			return -EINVAL;
 	}
 	if (!PKCS12_parse(p12, vpninfo->cert_password?:pass, &pkey, &cert, &ca)) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Parse PKCS#12 failed\n");
+		unsigned long err = ERR_peek_error();
+
 		report_ssl_errors(vpninfo);
+
+		if (ERR_GET_LIB(err) == ERR_LIB_PKCS12 &&
+		    ERR_GET_FUNC(err) == PKCS12_F_PKCS12_PARSE &&
+		    ERR_GET_REASON(err) == PKCS12_R_MAC_VERIFY_FAILURE) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Parse PKCS#12 failed (wrong passphrase?)\n");
+			vpninfo->cert_password = NULL;
+			goto retrypass;
+		}
+
+		vpninfo->progress(vpninfo, PRG_ERR, "Parse PKCS#12 failed (see above errors)\n");
 		PKCS12_free(p12);
 		return -EINVAL;
 	}
@@ -187,12 +206,13 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 						  buf, sizeof(buf));
 				vpninfo->progress(vpninfo, PRG_DEBUG,
 						  "Extra cert from PKCS#12: '%s'\n", buf);
+				CRYPTO_add(&cert2->references, 1, CRYPTO_LOCK_X509);
 				SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert2);
 				cert = cert2;
 				goto next;
 			}
 		}
-		sk_X509_free(ca);
+		sk_X509_pop_free(ca, X509_free);
 	}
 
 	PKCS12_free(p12);
@@ -333,7 +353,6 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					    SSL_FILETYPE_PEM)) {
 		unsigned long err = ERR_peek_error();
 		
-		vpninfo->progress(vpninfo, PRG_ERR, "Loading private key failed\n");
 		report_ssl_errors(vpninfo);
 
 #ifndef EVP_F_EVP_DECRYPTFINAL_EX
@@ -342,9 +361,12 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		/* If the user fat-fingered the passphrase, try again */
 		if (ERR_GET_LIB(err) == ERR_LIB_EVP &&
 		    ERR_GET_FUNC(err) == EVP_F_EVP_DECRYPTFINAL_EX &&
-		    ERR_GET_REASON(err) == EVP_R_BAD_DECRYPT)
+		    ERR_GET_REASON(err) == EVP_R_BAD_DECRYPT) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Loading private key failed (wrong passphrase?)\n");
 			goto again;
+		}
 		
+		vpninfo->progress(vpninfo, PRG_ERR, "Loading private key failed (see above errors)\n");
 		return -EINVAL;
 	}
 	return 0;
@@ -411,35 +433,310 @@ static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
 	return 0;
 }
 
-static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
+static int match_hostname_elem(const char *hostname, int helem_len,
+			       const char *match, int melem_len)
 {
-	X509 *peer_cert;
-
-	if (vpninfo->cafile) {
-		int vfy = SSL_get_verify_result(https_ssl);
-
-		if (vfy != X509_V_OK) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Server certificate verify failed: %s\n",
-				X509_verify_cert_error_string(vfy));
-			return -EINVAL;
-		}
+	if (!helem_len && !melem_len)
 		return 0;
+
+	if (!helem_len || !melem_len)
+		return -1;
+
+
+	if (match[0] == '*') {
+		int i;
+
+		for (i = 1 ; i <= helem_len; i++) {
+			if (!match_hostname_elem(hostname + i, helem_len - i,
+						 match + 1, melem_len - 1))
+				return 0;
+		}
+		return -1;
 	}
 
-	peer_cert = SSL_get_peer_certificate(https_ssl);
+	if (toupper(hostname[0]) == toupper(match[0]))
+		return match_hostname_elem(hostname + 1, helem_len - 1,
+					   match + 1, melem_len - 1);
 
-	if (vpninfo->servercert)
-		return check_server_cert(vpninfo, peer_cert);
+	return -1;
+}
 
-	if (vpninfo->validate_peer_cert)
-		return vpninfo->validate_peer_cert(vpninfo, peer_cert);
+static int match_hostname(const char *hostname, const char *match)
+{
+	while (*match) {
+		const char *h_dot, *m_dot;
+		int helem_len, melem_len;
+
+		h_dot = strchr(hostname, '.');
+		m_dot = strchr(match, '.');
+		
+		if (h_dot && m_dot) {
+			helem_len = h_dot - hostname + 1;
+			melem_len = m_dot - match + 1;
+		} else if (!h_dot && !m_dot) {
+			helem_len = strlen(hostname);
+			melem_len = strlen(match);
+		} else
+			return -1;
+
+
+		if (match_hostname_elem(hostname, helem_len,
+					match, melem_len))
+			return -1;
+
+		hostname += helem_len;
+		match += melem_len;
+	}
+	if (*hostname)
+		return -1;
 
-	/* If no validation function, just let it succeed */
 	return 0;
 }
 
-void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
-				      SSL *ssl)
+/* cf. RFC2818 and RFC2459 */
+int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
+{
+	STACK_OF(GENERAL_NAME) *altnames;
+	X509_NAME *subjname;
+	ASN1_STRING *subjasn1;
+	char *subjstr = NULL;
+	int addrlen = 0;
+	int i, altdns = 0;
+	char addrbuf[sizeof(struct in6_addr)];
+	int ret;
+
+	/* Allow GEN_IP in the certificate only if we actually connected
+	   by IP address rather than by name. */
+	if (inet_pton(AF_INET, vpninfo->hostname, addrbuf) > 0)
+		addrlen = 4;
+	else if (inet_pton(AF_INET6, vpninfo->hostname, addrbuf) > 0)
+		addrlen = 16;
+	else if (vpninfo->hostname[0] == '[' &&
+		 vpninfo->hostname[strlen(vpninfo->hostname)-1] == ']') {
+		char *p = &vpninfo->hostname[strlen(vpninfo->hostname)-1];
+		*p = 0;
+		if (inet_pton(AF_INET6, vpninfo->hostname + 1, addrbuf) > 0)
+			addrlen = 16;
+		*p = ']';
+	}
+
+	altnames = X509_get_ext_d2i(peer_cert, NID_subject_alt_name,
+				    NULL, NULL);
+	for (i = 0; i < sk_GENERAL_NAME_num(altnames); i++) {
+		const GENERAL_NAME *this = sk_GENERAL_NAME_value(altnames, i);
+
+		if (this->type == GEN_DNS) {
+			char *str;
+
+			int len = ASN1_STRING_to_UTF8((void *)&str, this->d.ia5);
+			if (len < 0)
+				continue;
+
+			altdns = 1;
+
+			/* We don't like names with embedded NUL */
+			if (strlen(str) != len)
+				continue;
+
+			if (!match_hostname(vpninfo->hostname, str)) {
+				vpninfo->progress(vpninfo, PRG_TRACE,
+						  "Matched DNS altname '%s'\n",
+						  str);
+				GENERAL_NAMES_free(altnames);
+				OPENSSL_free(str);
+				return 0;
+			} else {
+				vpninfo->progress(vpninfo, PRG_TRACE,
+						  "No match for altname '%s'\n",
+						  str);
+			}
+			OPENSSL_free(str);
+		} else if (this->type == GEN_IPADD && addrlen) {
+			char host[80];
+			int family;
+
+			if (this->d.ip->length == 4) {
+				family = AF_INET;
+			} else if (this->d.ip->length == 16) {
+				family = AF_INET6;
+			} else {
+				vpninfo->progress(vpninfo, PRG_ERR,
+						  "Certificate has GEN_IPADD altname with bogus length %d\n",
+						  this->d.ip->length);
+				continue;
+			}
+			
+			/* We only do this for the debug messages */
+			inet_ntop(family, this->d.ip->data, host, sizeof(host));
+
+			if (this->d.ip->length == addrlen &&
+			    !memcmp(addrbuf, this->d.ip->data, addrlen)) {
+				vpninfo->progress(vpninfo, PRG_TRACE,
+						  "Matched IP%s address '%s'\n",
+						  (family == AF_INET6)?"v6":"",
+						  host);
+				GENERAL_NAMES_free(altnames);
+				return 0;
+			} else {
+				vpninfo->progress(vpninfo, PRG_TRACE,
+						  "No match for IP%s address '%s'\n",
+						  (family == AF_INET6)?"v6":"",
+						  host);
+			}
+		} else if (this->type == GEN_URI) {
+			char *str;
+			char *url_proto, *url_host, *url_path, *url_host2;
+			int url_port;
+			int len = ASN1_STRING_to_UTF8((void *)&str, this->d.ia5);
+
+			if (len < 0)
+				continue;
+
+			/* We don't like names with embedded NUL */
+			if (strlen(str) != len)
+				continue;
+
+			if (parse_url(str, &url_proto, &url_host, &url_port, &url_path, 0)) {
+				OPENSSL_free(str);
+				continue;
+			}
+
+			if (!url_proto || strcasecmp(url_proto, "https"))
+				goto no_uri_match;
+
+			if (url_port != vpninfo->port)
+				goto no_uri_match;
+
+			/* Leave url_host as it was so that it can be freed */
+			url_host2 = url_host;
+			if (addrlen == 16 && vpninfo->hostname[0] != '[' &&
+			    url_host[0] == '[' && url_host[strlen(url_host)-1] == ']') {
+				/* Cope with https://[IPv6]/ when the hostname is bare IPv6 */
+				url_host[strlen(url_host)-1] = 0;
+				url_host2++;
+			}
+
+			if (strcasecmp(vpninfo->hostname, url_host2))
+				goto no_uri_match;
+
+			if (url_path) {
+				vpninfo->progress(vpninfo, PRG_TRACE, "URI '%s' has non-empty path; ignoring\n",
+						  str);
+				goto no_uri_match_silent;
+			}
+			vpninfo->progress(vpninfo, PRG_TRACE,
+					  "Matched URI '%s'\n",
+					  str);
+			free(url_proto);
+			free(url_host);
+			free(url_path);
+			OPENSSL_free(str);
+			GENERAL_NAMES_free(altnames);
+			return 0;
+
+		no_uri_match:
+			vpninfo->progress(vpninfo, PRG_TRACE,
+					  "No match for URI '%s'\n",
+					  str);
+		no_uri_match_silent:
+			free(url_proto);
+			free(url_host);
+			free(url_path);
+			OPENSSL_free(str);
+		}
+	}
+	GENERAL_NAMES_free(altnames);
+
+	/* According to RFC2818, we don't use the legacy subject name if
+	   there was an altname with DNS type. */
+	if (altdns) {
+		vpninfo->progress(vpninfo, PRG_ERR, "No altname in peer cert matched '%s'\n",
+				  vpninfo->hostname);
+		return -EINVAL;
+	}
+
+	subjname = X509_get_subject_name(peer_cert);
+	if (!subjname) {
+		vpninfo->progress(vpninfo, PRG_ERR, "No subject name in peer cert!\n");
+		return -EINVAL;
+	}
+
+	/* Find the _last_ (most specific) commonName */
+	i = -1;
+	while (1) {
+		int j = X509_NAME_get_index_by_NID(subjname, NID_commonName, i);
+		if (j >= 0)
+			i = j;
+		else
+			break;
+	}
+
+	subjasn1 = X509_NAME_ENTRY_get_data(X509_NAME_get_entry(subjname, i));
+
+	i = ASN1_STRING_to_UTF8((void *)&subjstr, subjasn1);
+
+	if (!subjstr || strlen(subjstr) != i) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Failed to parse subject name in peer cert\n");
+		return -EINVAL;
+	}
+	ret = 0;
+
+	if (match_hostname(vpninfo->hostname, subjstr)) {
+		vpninfo->progress(vpninfo, PRG_ERR, "Peer cert subject mismatch ('%s' != '%s')\n",
+				  subjstr, vpninfo->hostname);
+		ret = -EINVAL;
+	} else {
+		vpninfo->progress(vpninfo, PRG_TRACE,
+				  "Matched peer certificate subject name '%s'\n",
+				  subjstr);
+	}
+
+	OPENSSL_free(subjstr);			  
+	return ret;
+}
+
+static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
+{
+	X509 *peer_cert;
+	int ret;
+
+	peer_cert = SSL_get_peer_certificate(https_ssl);
+
+	if (vpninfo->servercert) {
+		/* If given a cert fingerprint on the command line, that's
+		   all we look for */
+		ret = check_server_cert(vpninfo, peer_cert);
+	} else {
+		int vfy = SSL_get_verify_result(https_ssl);
+		const char *err_string = NULL;
+
+		if (vfy != X509_V_OK)
+			err_string = X509_verify_cert_error_string(vfy);
+		else if (match_cert_hostname(vpninfo, peer_cert))
+			err_string = "certificate does not match hostname";
+
+		if (err_string) {
+			vpninfo->progress(vpninfo, PRG_ERR,
+					  "Server certificate verify failed: %s\n",
+					  err_string);
+
+			if (vpninfo->validate_peer_cert)
+				ret = vpninfo->validate_peer_cert(vpninfo, peer_cert,
+								  err_string);
+			else
+				ret = -EINVAL;
+		} else {
+			ret = 0;
+		}
+	}
+	X509_free(peer_cert);
+
+	return ret;
+}
+
+static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
+					     SSL *ssl)
 {
 	/* OpenSSL has problems with certificate chains -- if there are
 	   multiple certs with the same name, it doesn't necessarily
@@ -474,6 +771,16 @@ void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 	X509_STORE_CTX_cleanup(&ctx);
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x00908000
+static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
+{
+	/* We've seen certificates in the wild which don't have the
+	   purpose fields filled in correctly */
+	X509_VERIFY_PARAM_set_purpose(ctx->param, X509_PURPOSE_ANY);
+	return X509_verify_cert(ctx);
+}
+#endif
+
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
 	method_const SSL_METHOD *ssl3_method;
@@ -581,8 +888,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			free(hostname);
 
 		if (err) {
-			vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed: %s\n",
-					  gai_strerror(err));
+			vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed for host '%s': %s\n",
+					  hostname, gai_strerror(err));
 			return -EINVAL;
 		}
 
@@ -649,9 +956,18 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			}
 		}
 
-		/* We've seen certificates in the wild which don't have the
-		   purpose fields filled in correctly */
-		SSL_CTX_set_purpose(vpninfo->https_ctx, X509_PURPOSE_ANY);
+		/* We just want to do:
+		   SSL_CTX_set_purpose(vpninfo->https_ctx, X509_PURPOSE_ANY); 
+		   ... but it doesn't work with OpenSSL < 0.9.8k because of 
+		   problems with inheritance (fixed in v1.1.4.6 of
+		   crypto/x509/x509_vpm.c) so we have to play silly buggers
+		   instead. This trick doesn't work _either_ in < 0.9.7 but
+		   I don't know of _any_ workaround which will, and can't
+		   be bothered to find out either. */
+#if OPENSSL_VERSION_NUMBER >= 0x00908000
+		SSL_CTX_set_cert_verify_callback(vpninfo->https_ctx,
+						 ssl_app_verify_callback, NULL);
+#endif
 		SSL_CTX_set_default_verify_paths(vpninfo->https_ctx);
 
 		if (vpninfo->cafile)
@@ -709,7 +1025,7 @@ void openconnect_init_openssl(void)
 	OpenSSL_add_all_algorithms ();
 }
 
-#ifdef __sun__
+#if defined(__sun__) || defined(__NetBSD__) || defined(__DragonFly__)
 int passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	struct statvfs buf;
